import * as i0 from '@angular/core';
import { Injectable, InjectionToken, NgModule } from '@angular/core';
import 'reflect-metadata';
import { Observable, Subject, from } from 'rxjs';
import { mergeMap, filter } from 'rxjs/operators';
import { v4 } from 'uuid';

var _a, _b;
const INTERNAL_EVENTS = Symbol();
const IS_AUTO_COMMIT_ENABLED = Symbol();
class AggregateRoot {
    constructor() {
        this[_a] = false;
        this[_b] = [];
    }
    set autoCommit(value) {
        this[IS_AUTO_COMMIT_ENABLED] = value;
    }
    get autoCommit() {
        return this[IS_AUTO_COMMIT_ENABLED];
    }
    hangup() { }
    resume(data) { }
    publish(event) { }
    publishAll(event) { }
    commit() {
        this.publishAll(this[INTERNAL_EVENTS]);
        this[INTERNAL_EVENTS].length = 0;
    }
    uncommit() {
        this[INTERNAL_EVENTS].length = 0;
    }
    getUncommittedEvents() {
        return this[INTERNAL_EVENTS];
    }
    loadFromHistory(history) {
        history.forEach((event) => this.apply(event, true));
    }
    apply(event, isFromHistory = false) {
        if (!isFromHistory && !this.autoCommit) {
            this[INTERNAL_EVENTS].push(event);
        }
        this.autoCommit && this.publish(event);
        const handler = this.getEventHandler(event);
        handler && handler.call(this, event);
    }
    getEventHandler(event) {
        const handler = `on${this.getEventName(event)}`;
        return this[handler];
    }
    getEventName(event) {
        const { constructor } = Object.getPrototypeOf(event);
        return constructor.name;
    }
}
_a = IS_AUTO_COMMIT_ENABLED, _b = INTERNAL_EVENTS;

const COMMAND_METADATA = '__command__';
const COMMAND_HANDLER_METADATA = '__commandHandler__';
const QUERY_METADATA = '__query__';
const QUERY_HANDLER_METADATA = '__queryHandler__';
const EVENT_METADATA = '__event__';
const EVENTS_HANDLER_METADATA = '__eventsHandler__';
const SAGA_METADATA = '__saga__';

class CommandHandlerNotFoundException extends Error {
    constructor(commandName) {
        super(`The command handler for the "${commandName}" command was not found!`);
    }
}

class DefaultCommandPubSub {
    constructor(subject$) {
        this.subject$ = subject$;
    }
    publish(command) {
        this.subject$.next(command);
    }
}

class ObservableBus extends Observable {
    constructor() {
        super();
        this._subject$ = new Subject();
        this.source = this._subject$;
    }
    get subject$() {
        return this._subject$;
    }
}

class CommandBus extends ObservableBus {
    constructor(moduleRef) {
        super();
        this.moduleRef = moduleRef;
        this.handlers = new Map();
        this.useDefaultPublisher();
    }
    get publisher() {
        return this._publisher;
    }
    set publisher(_publisher) {
        this._publisher = _publisher;
    }
    execute(command) {
        const commandId = this.getCommandId(command);
        const handler = this.handlers.get(commandId);
        if (!handler) {
            throw new CommandHandlerNotFoundException(commandId);
        }
        this._publisher.publish(command);
        return handler.execute(command);
    }
    bind(handler, id) {
        this.handlers.set(id, handler);
    }
    register(handlers = []) {
        handlers.forEach((handler) => this.registerHandler(handler));
    }
    registerHandler(handler) {
        this.moduleRef.injector.runInContext(() => {
            this.moduleRef.injector.get(handler);
            const instance = this.moduleRef.injector.get(handler);
            if (!instance) {
                return;
            }
            const target = this.reflectCommandId(handler);
            if (!target) {
                throw new InvalidCommandHandlerException();
            }
            this.bind(instance, target);
        });
    }
    getCommandId(command) {
        const { constructor: commandType } = Object.getPrototypeOf(command);
        const commandMetadata = Reflect.getMetadata(COMMAND_METADATA, commandType);
        if (!commandMetadata) {
            throw new CommandHandlerNotFoundException(commandType.name);
        }
        return commandMetadata.id;
    }
    reflectCommandId(handler) {
        const command = Reflect.getMetadata(COMMAND_HANDLER_METADATA, handler);
        const commandMetadata = Reflect.getMetadata(COMMAND_METADATA, command);
        return commandMetadata.id;
    }
    useDefaultPublisher() {
        this._publisher = new DefaultCommandPubSub(this.subject$);
    }
}
CommandBus.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: CommandBus, deps: [{ token: i0.NgModuleRef }], target: i0.ɵɵFactoryTarget.Injectable });
CommandBus.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: CommandBus, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: CommandBus, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: i0.NgModuleRef }]; } });

class InvalidCommandHandlerException extends Error {
    constructor() {
        super(`Invalid command handler exception (missing @CommandHandler() decorator?)`);
    }
}

class InvalidEventsHandlerException extends Error {
    constructor() {
        super(`Invalid event handler exception (missing @EventsHandler() decorator?)`);
    }
}

class InvalidQueryHandlerException extends Error {
    constructor() {
        super(`Invalid query handler exception (missing @QueryHandler() decorator?)`);
    }
}

class InvalidSagaException extends Error {
    constructor() {
        super(`Invalid saga exception. Each saga should return an Observable object`);
    }
}

class QueryHandlerNotFoundException extends Error {
    constructor(queryName) {
        super(`The query handler for the "${queryName}" query was not found!`);
    }
}

/**
 * Null if the published class is not connected to any handler
 * @param event
 * @returns
 */
const defaultGetEventId = (event) => {
    const { constructor } = Object.getPrototypeOf(event);
    return Reflect.getMetadata(EVENT_METADATA, constructor)?.id ?? null;
};
const defaultReflectEventId = (event) => {
    return Reflect.getMetadata(EVENT_METADATA, event).id;
};

class DefaultPubSub {
    constructor(subject$) {
        this.subject$ = subject$;
    }
    publish(event) {
        this.subject$.next(event);
    }
    bridgeEventsTo(subject) {
        this.subject$ = subject;
    }
}

class EventBus extends ObservableBus {
    // private readonly _logger = new Logger(EventBus.name);
    constructor(commandBus, moduleRef) {
        super();
        this.commandBus = commandBus;
        this.moduleRef = moduleRef;
        this.subscriptions = [];
        this.getEventId = defaultGetEventId;
        this.useDefaultPublisher();
    }
    get publisher() {
        return this._publisher;
    }
    set publisher(_publisher) {
        this._publisher = _publisher;
    }
    ngOnDestroy() {
        this.subscriptions.forEach((subscription) => subscription.unsubscribe());
    }
    publish(event) {
        return this._publisher.publish(event);
    }
    publishAll(events) {
        if (this._publisher.publishAll) {
            return this._publisher.publishAll(events);
        }
        return (events || []).map((event) => this._publisher.publish(event));
    }
    bind(handler, id) {
        const stream$ = id ? this.ofEventId(id) : this.subject$;
        const subscription = stream$
            .pipe(mergeMap((event) => from(Promise.resolve(handler.handle(event)))))
            .subscribe({
            error: (error) => {
                // this._logger.error(
                //   `"${handler.constructor.name}" has thrown an error.`,
                //   error,
                // );
                throw error;
            },
        });
        this.subscriptions.push(subscription);
    }
    registerSagas(types = []) {
        const sagas = types
            .map((target) => {
            const metadata = Reflect.getMetadata(SAGA_METADATA, target) || [];
            const instance = this.moduleRef.injector.get(target);
            if (!instance) {
                throw new InvalidSagaException();
            }
            return metadata.map((key) => instance[key].bind(instance));
        })
            .reduce((a, b) => a.concat(b), []);
        sagas.forEach((saga) => this.registerSaga(saga));
    }
    register(handlers = []) {
        handlers.forEach((handler) => this.registerHandler(handler));
    }
    registerHandler(handler) {
        const instance = this.moduleRef.injector.get(handler);
        if (!instance) {
            return;
        }
        const events = this.reflectEvents(handler);
        events.map((event) => this.bind(instance, defaultReflectEventId(event)));
    }
    ofEventId(id) {
        return this.subject$.pipe(filter((event) => this.getEventId(event) === id));
    }
    registerSaga(saga) {
        if (typeof saga !== 'function') {
            throw new InvalidSagaException();
        }
        const stream$ = saga(this);
        if (!(stream$ instanceof Observable)) {
            throw new InvalidSagaException();
        }
        const subscription = stream$
            .pipe(filter((e) => !!e), mergeMap((command) => from(this.commandBus.execute(command))))
            .subscribe({
            error: (error) => {
                // this._logger.error(
                //   `Command handler which execution was triggered by Saga has thrown an error.`,
                //   error,
                // );
                throw error;
            },
        });
        this.subscriptions.push(subscription);
    }
    reflectEvents(handler) {
        return Reflect.getMetadata(EVENTS_HANDLER_METADATA, handler);
    }
    useDefaultPublisher() {
        this._publisher = new DefaultPubSub(this.subject$);
    }
}
EventBus.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: EventBus, deps: [{ token: CommandBus }, { token: i0.NgModuleRef }], target: i0.ɵɵFactoryTarget.Injectable });
EventBus.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: EventBus, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: EventBus, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: CommandBus }, { type: i0.NgModuleRef }]; } });

class EventPublisher {
    constructor(eventBus) {
        this.eventBus = eventBus;
    }
    mergeClassContext(metatype) {
        const eventBus = this.eventBus;
        return class extends metatype {
            publish(event) {
                eventBus.publish(event);
            }
            publishAll(events) {
                eventBus.publishAll(events);
            }
        };
    }
    mergeObjectContext(object) {
        const eventBus = this.eventBus;
        object.publish = (event) => {
            eventBus.publish(event);
        };
        object.publishAll = (events) => {
            eventBus.publishAll(events);
        };
        return object;
    }
}
EventPublisher.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: EventPublisher, deps: [{ token: EventBus }], target: i0.ɵɵFactoryTarget.Injectable });
EventPublisher.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: EventPublisher, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: EventPublisher, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: EventBus }]; } });

class DefaultQueryPubSub {
    constructor(subject$) {
        this.subject$ = subject$;
    }
    publish(query) {
        this.subject$.next(query);
    }
}

class QueryBus extends ObservableBus {
    constructor(moduleRef) {
        super();
        this.moduleRef = moduleRef;
        this.handlers = new Map();
        this.useDefaultPublisher();
    }
    get publisher() {
        return this._publisher;
    }
    set publisher(_publisher) {
        this._publisher = _publisher;
    }
    execute(query) {
        const queryId = this.getQueryId(query);
        const handler = this.handlers.get(queryId);
        if (!handler) {
            throw new QueryHandlerNotFoundException(queryId);
        }
        this._publisher.publish(query);
        return handler.execute(query);
    }
    bind(handler, queryId) {
        this.handlers.set(queryId, handler);
    }
    register(handlers = []) {
        handlers.forEach((handler) => this.registerHandler(handler));
    }
    registerHandler(handler) {
        const instance = this.moduleRef.injector.get(handler);
        if (!instance) {
            return;
        }
        const target = this.reflectQueryId(handler);
        if (!target) {
            throw new InvalidQueryHandlerException();
        }
        this.bind(instance, target);
    }
    getQueryId(query) {
        const { constructor: queryType } = Object.getPrototypeOf(query);
        const queryMetadata = Reflect.getMetadata(QUERY_METADATA, queryType);
        if (!queryMetadata) {
            throw new QueryHandlerNotFoundException(queryType.name);
        }
        return queryMetadata.id;
    }
    reflectQueryId(handler) {
        const query = Reflect.getMetadata(QUERY_HANDLER_METADATA, handler);
        const queryMetadata = Reflect.getMetadata(QUERY_METADATA, query);
        return queryMetadata.id;
    }
    useDefaultPublisher() {
        this._publisher = new DefaultQueryPubSub(this.subject$);
    }
}
QueryBus.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: QueryBus, deps: [{ token: i0.NgModuleRef }], target: i0.ɵɵFactoryTarget.Injectable });
QueryBus.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: QueryBus, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: QueryBus, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: i0.NgModuleRef }]; } });

const PROVIDERS = new InjectionToken('providers');
class BindService {
    constructor(moduleRef, eventsBus, commandsBus, queryBus) {
        this.moduleRef = moduleRef;
        this.eventsBus = eventsBus;
        this.commandsBus = commandsBus;
        this.queryBus = queryBus;
        this._explored = false;
    }
    get explored() {
        return this._explored;
    }
    bind() {
        const injector = this.moduleRef.injector;
        const providers = injector.get(PROVIDERS);
        const commands = this.flatMap(providers, (instance) => this.filterProvider(instance, COMMAND_HANDLER_METADATA));
        const queries = this.flatMap(providers, (instance) => this.filterProvider(instance, QUERY_HANDLER_METADATA));
        const events = this.flatMap(providers, (instance) => this.filterProvider(instance, EVENTS_HANDLER_METADATA));
        const sagas = this.flatMap(providers, (instance) => this.filterProvider(instance, SAGA_METADATA));
        this.eventsBus.register(events);
        this.commandsBus.register(commands);
        this.queryBus.register(queries);
        this.eventsBus.registerSagas(sagas);
        this._explored = true;
    }
    flatMap(providers, callback) {
        return providers
            .map(callback)
            .reduce((a, b) => a.concat(b), [])
            .filter((element) => !!element);
    }
    filterProvider(instance, metadataKey) {
        if (!instance) {
            return undefined;
        }
        return this.extractMetadata(instance, metadataKey);
    }
    extractMetadata(instance, metadataKey) {
        if (!instance.constructor) {
            return null;
        }
        const metadata = Reflect.getMetadata(metadataKey, instance);
        return metadata ? instance : undefined;
    }
}
BindService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: BindService, deps: [{ token: i0.NgModuleRef }, { token: EventBus }, { token: CommandBus }, { token: QueryBus }], target: i0.ɵɵFactoryTarget.Injectable });
BindService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: BindService, providedIn: 'root' });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: BindService, decorators: [{
            type: Injectable,
            args: [{ providedIn: 'root' }]
        }], ctorParameters: function () { return [{ type: i0.NgModuleRef }, { type: EventBus }, { type: CommandBus }, { type: QueryBus }]; } });

class CqrsModule {
    constructor(bindService) {
        bindService.bind();
    }
}
CqrsModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: CqrsModule, deps: [{ token: BindService }], target: i0.ɵɵFactoryTarget.NgModule });
CqrsModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "14.0.0", version: "15.0.4", ngImport: i0, type: CqrsModule });
CqrsModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: CqrsModule, providers: [CommandBus, QueryBus, EventBus, EventPublisher, BindService] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.4", ngImport: i0, type: CqrsModule, decorators: [{
            type: NgModule,
            args: [{
                    providers: [CommandBus, QueryBus, EventBus, EventPublisher, BindService],
                }]
        }], ctorParameters: function () { return [{ type: BindService }]; } });

/**
 * Decorator that marks a class as a Nest command handler. A command handler
 * handles commands (actions) executed by your application code.
 *
 * The decorated class must implement the `ICommandHandler` interface.
 *
 * @param command command *type* to be handled by this handler.
 *
 * @see https://docs.nestjs.com/recipes/cqrs#commands
 */
const CommandHandler = (command) => {
    return (target) => {
        if (!Reflect.hasMetadata(COMMAND_METADATA, command)) {
            Reflect.defineMetadata(COMMAND_METADATA, { id: v4() }, command);
        }
        Reflect.defineMetadata(COMMAND_HANDLER_METADATA, command, target);
    };
};

/**
 * Decorator that marks a class as a Nest event handler. An event handler
 * handles events executed by your application code.
 *
 * The decorated class must implement the `IEventHandler` interface.
 *
 * @param events one or more event *types* to be handled by this handler.
 *
 * @see https://docs.nestjs.com/recipes/cqrs#events
 */
const EventsHandler = (...events) => {
    return (target) => {
        events.forEach((event) => {
            if (!Reflect.hasOwnMetadata(EVENT_METADATA, event)) {
                Reflect.defineMetadata(EVENT_METADATA, { id: v4() }, event);
            }
        });
        Reflect.defineMetadata(EVENTS_HANDLER_METADATA, events, target);
    };
};

/**
 * Decorator that marks a class as a Nest query handler. A query handler
 * handles queries executed by your application code.
 *
 * The decorated class must implement the `IQueryHandler` interface.
 *
 * @param query query *type* to be handled by this handler.
 *
 * @see https://docs.nestjs.com/recipes/cqrs#queries
 */
const QueryHandler = (query) => {
    return (target) => {
        if (!Reflect.hasMetadata(QUERY_METADATA, query)) {
            Reflect.defineMetadata(QUERY_METADATA, { id: v4() }, query);
        }
        Reflect.defineMetadata(QUERY_HANDLER_METADATA, query, target);
    };
};

/**
 * Decorator that marks a class as a Nest saga. Sagas may listen and react to 1..N events.
 *
 * @see https://docs.nestjs.com/recipes/cqrs#sagas
 */
const Saga = () => {
    return (target, propertyKey) => {
        const properties = Reflect.getMetadata(SAGA_METADATA, target.constructor) || [];
        Reflect.defineMetadata(SAGA_METADATA, [...properties, propertyKey], target.constructor);
    };
};

/**
 * Filter values depending on their instance type (comparison is made
 * using native `instanceof`).
 *
 * @param types List of types implementing `IEvent`.
 *
 * @return A stream only emitting the filtered instances.
 */
function ofType(...types) {
    const isInstanceOf = (event) => !!types.find((classType) => event instanceof classType);
    return (source) => source.pipe(filter(isInstanceOf));
}

/*
 * Nest CQRS Module
 * Copyright(c) 2017-... Kamil Mysliwiec
 * www.kamilmysliwiec.com
 * MIT Licensed
 */

/**
 * Generated bundle index. Do not edit.
 */

export { AggregateRoot, BindService, CommandBus, CommandHandler, CommandHandlerNotFoundException, CqrsModule, EventBus, EventPublisher, EventsHandler, InvalidCommandHandlerException, InvalidEventsHandlerException, InvalidQueryHandlerException, InvalidSagaException, ObservableBus, PROVIDERS, QueryBus, QueryHandler, QueryHandlerNotFoundException, Saga, ofType };
//# sourceMappingURL=cqrs.mjs.map
//# sourceMappingURL=cqrs.mjs.map
